diff --git a/lib/ssh/src/ssh_connection_handler.erl b/lib/ssh/src/ssh_connection_handler.erl
index 59e80e6..dedb2e6 100644
--- a/lib/ssh/src/ssh_connection_handler.erl
+++ b/lib/ssh/src/ssh_connection_handler.erl
@@ -69,7 +69,8 @@
 	  key_exchange_init_msg,   %  #ssh_msg_kexinit{}
 	  renegotiate = false,     %  boolean() 
 	  last_size_rekey = 0,
-	  connection_queue,
+	  kexinit_reply_queue = [],
+	  kexinit_event_queue = [],
 	  address,
 	  port,
 	  opts
@@ -379,9 +380,12 @@ key_exchange(#ssh_msg_kex_dh_gex_reply{} = Msg,
 
 new_keys(#ssh_msg_newkeys{} = Msg, #state{ssh_params = Ssh0} = State0) ->
     {ok, Ssh} = ssh_transport:handle_new_keys(Msg, Ssh0),
-    {NextStateName, State} =
-	after_new_keys(State0#state{ssh_params = Ssh}),
-    {next_state, NextStateName, next_packet(State)}.
+    case after_new_keys(State0#state{ssh_params = Ssh}) of
+	{stop, _Reason, _NewStateData} = Terminator ->
+	    Terminator;
+	{NextStateName, State} ->
+	    {next_state, NextStateName, next_packet(State)}
+    end.
 
 %%--------------------------------------------------------------------
 -spec userauth(#ssh_msg_service_request{} | #ssh_msg_service_accept{} |
@@ -675,6 +679,11 @@ handle_sync_event({global_request, Pid, _, _, _} = Request, From, StateName,
     State = add_request(true, Channel#channel.local_id, From, State1),
     {next_state, StateName, next_packet(State)};
 
+handle_sync_event({data, ChannelId, Type, Data, Timeout}, From, StateName, State)
+  when StateName == kexinit orelse State#state.renegotiate == true ->
+    Events = [{{data, ChannelId, Type, Data, Timeout}, From} | State#state.kexinit_event_queue],
+    {next_state, StateName, State#state{kexinit_event_queue = Events}};
+
 handle_sync_event({data, ChannelId, Type, Data, Timeout}, From, StateName,
 		  #state{connection_state = #connection{channel_cache = _Cache}
 			 = Connection0} = State0) ->
@@ -1185,7 +1194,12 @@ generate_event(<<?BYTE(Byte), _/binary>> = Msg, StateName,
     State1 = generate_event_new_state(State0, EncData),
     try ssh_connection:handle_msg(ConnectionMsg, Connection0, Role) of
 	{{replies, Replies}, Connection} ->
-	    State = send_replies(Replies,  State1#state{connection_state = Connection}),
+	    if Renegotiation == true ->
+		    NewReplies = State1#state.kexinit_reply_queue ++ Replies,
+		    State = State1#state{connection_state = Connection, kexinit_reply_queue = NewReplies };
+	       true ->
+		    State = send_replies(Replies,  State1#state{connection_state = Connection})
+	    end,
 	    {next_state, StateName, next_packet(State)};
 	{noreply, Connection} ->
 	    {next_state, StateName, next_packet(State1#state{connection_state = Connection})};
@@ -1348,7 +1362,11 @@ next_packet(#state{socket = Socket} = State) ->
     State.
 
 after_new_keys(#state{renegotiate = true} = State) ->
-    {connected, State#state{renegotiate = false}};
+    State1 = State#state{renegotiate = false},
+    {NewStateName, State2} =
+	lists:foldr(fun after_new_keys_sync_event/2, {connected, State1}, State#state.kexinit_event_queue),
+    State3 = send_replies(State#state.kexinit_reply_queue, State2),
+    {NewStateName, State3#state{kexinit_reply_queue = [], kexinit_event_queue = []}};
 after_new_keys(#state{renegotiate = false, 
 		      ssh_params = #ssh{role = client} = Ssh0} = State) ->
     {Msg, Ssh} = ssh_auth:service_request_msg(Ssh0),
@@ -1358,6 +1376,21 @@ after_new_keys(#state{renegotiate = false,
 		      ssh_params = #ssh{role = server}} = State) ->
     {userauth, State}.
 
+after_new_keys_sync_event({_Event, From}, {stop, _Reason, _StateData}=Terminator) ->
+    gen_fsm:reply(From, {error, closed}),
+    Terminator;
+after_new_keys_sync_event({Event, From}, {StateName, StateData}) ->
+    case handle_sync_event(Event, From, StateName, StateData) of
+	{reply, Reply, NextStateName, NewStateData} ->
+	    gen_fsm:reply(From, Reply),
+	    {NextStateName, NewStateData};
+	{next_state, NextStateName, NewStateData}->
+	    {NextStateName, NewStateData};
+	{stop, Reason, Reply, NewStateData} ->
+	    gen_fsm:reply(From, Reply),
+	    {stop, Reason, NewStateData}
+    end.
+
 handle_ssh_packet_data(RemainingSshPacketLen, DecData, EncData, StateName, 
 		       State) ->
     EncSize =  size(EncData), 
